@@ -49,6 +49,8 @@ impl<'a> PartialEq for OrdVertex<'a> {
 }
 impl<'a> Eq for OrdVertex<'a> {}
 
+type PredecessorInfo<'a> = (u64, Vertex<'a>, Edge);
+
 fn shortest_path(start: Vertex) -> Vec<(Edge, Vertex)> {
     // We want to visit nodes with the shortest distance first, but
     // BinaryHeap is a max-heap. Ensure nodes are wrapped with Reverse
@@ -59,39 +61,41 @@ fn shortest_path(start: Vertex) -> Vec<(Edge, Vertex)> {
 
     // TODO: this grows very big. Consider using IDA* to reduce memory
     // usage.
-    let mut predecessors: FxHashMap<Vertex, (u64, Vertex, Edge)> = FxHashMap::default();
+    let mut predecessors: FxHashMap<Vertex, PredecessorInfo> = FxHashMap::default();
 
-    let end;
-    loop {
+    let mut neighbour_buf = [None, None, None, None, None, None, None, None, None, None];
+    let end = loop {
         match heap.pop() {
             Some((Reverse(distance), current)) => {
                 if current.is_end() {
-                    end = current;
-                    break;
+                    break current;
                 }
 
-                for (edge, next) in neighbours(&current) {
+                neighbours(&current, &mut neighbour_buf);
+                for (edge, next) in neighbour_buf.iter().flatten() {
                     let distance_to_next = distance + edge.cost();
-                    let found_shorter_route = match predecessors.get(&next) {
+                    let found_shorter_route = match predecessors.get(next) {
                         Some((prev_shortest, _, _)) => distance_to_next < *prev_shortest,
                         _ => true,
                     };
 
                     if found_shorter_route {
                         predecessors
-                            .insert(next.clone(), (distance_to_next, current.clone(), edge));
+                            .insert(next.clone(), (distance_to_next, current.clone(), *edge));
 
-                        heap.push(Reverse(distance_to_next), next);
+                        heap.push(Reverse(distance_to_next), next.clone());
                     }
                 }
             }
             None => panic!("Ran out of graph nodes before reaching end"),
         }
-    }
+    };
 
     info!(
-        "Found predecessors for {} syntax nodes, with {} left on heap.",
+        "Found predecessors for {} syntax nodes (hashmap key: {} bytes, value: {} bytes), with {} left on heap.",
         predecessors.len(),
+        std::mem::size_of::<Vertex>(),
+        std::mem::size_of::<PredecessorInfo>(),
         heap.len(),
     );
     let mut current = end;
@@ -127,11 +131,10 @@ mod tests {
     use crate::{
         graph::Edge::*,
         positions::SingleLineSpan,
-        syntax::{init_info, ChangeKind, Syntax::*, SyntaxInfo},
+        syntax::{init_info, ChangeKind},
     };
 
     use itertools::Itertools;
-    use std::cell::Cell;
     use typed_arena::Arena;
 
     fn pos_helper(line: usize) -> Vec<SingleLineSpan> {
@@ -154,26 +157,10 @@ mod tests {
     fn identical_atoms() {
         let arena = Arena::new();
 
-        let lhs = arena.alloc(Atom {
-            info: SyntaxInfo {
-                unique_id: Cell::new(0),
-                ..SyntaxInfo::new(0)
-            },
-            position: pos_helper(0),
-            content: "foo".into(),
-            is_comment: false,
-        });
-
+        let lhs = Syntax::new_atom(&arena, pos_helper(0), "foo");
         // Same content as LHS.
-        let rhs = arena.alloc(Atom {
-            info: SyntaxInfo {
-                unique_id: Cell::new(1),
-                ..SyntaxInfo::new(1)
-            },
-            position: pos_helper(1),
-            content: "foo".into(),
-            is_comment: false,
-        });
+        let rhs = Syntax::new_atom(&arena, pos_helper(0), "foo");
+        init_info(&[lhs], &[rhs]);
 
         let start = Vertex {
             lhs_syntax: Some(lhs),
@@ -639,8 +626,8 @@ mod tests {
         init_info(&[lhs], &[rhs]);
 
         mark_syntax(Some(lhs), Some(rhs));
-        assert_eq!(lhs.info().change.get(), Some(ChangeKind::Unchanged(rhs)));
-        assert_eq!(rhs.info().change.get(), Some(ChangeKind::Unchanged(lhs)));
+        assert_eq!(lhs.change(), Some(ChangeKind::Unchanged(rhs)));
+        assert_eq!(rhs.change(), Some(ChangeKind::Unchanged(lhs)));
     }
 
     #[test]
@@ -651,7 +638,7 @@ mod tests {
         init_info(&[lhs], &[rhs]);
 
         mark_syntax(Some(lhs), Some(rhs));
-        assert_eq!(lhs.info().change.get(), Some(ChangeKind::Novel));
-        assert_eq!(rhs.info().change.get(), Some(ChangeKind::Novel));
+        assert_eq!(lhs.change(), Some(ChangeKind::Novel));
+        assert_eq!(rhs.change(), Some(ChangeKind::Novel));
     }
 }
