@@ -6,7 +6,7 @@ use itertools::{EitherOrBoth, Itertools};
 use std::{
     cell::Cell,
     cmp::min,
-    collections::{hash_map::DefaultHasher, HashMap},
+    collections::HashMap,
     env, fmt,
     hash::{Hash, Hasher},
 };
@@ -42,26 +42,27 @@ impl<'a> fmt::Debug for ChangeKind<'a> {
 
 /// Fields that are common to both `Syntax::List` and `Syntax::Atom`.
 pub struct SyntaxInfo<'a> {
-    // TODO: Make these fields private.
-    pub pos_content_hash: u64,
-    pub next: Cell<Option<&'a Syntax<'a>>>,
-    pub prev: Cell<Option<&'a Syntax<'a>>>,
-    pub prev_is_contiguous: Cell<bool>,
-    pub change: Cell<Option<ChangeKind<'a>>>,
-    pub num_ancestors: Cell<u64>,
-    pub unique_id: Cell<u64>,
+    next: Cell<Option<&'a Syntax<'a>>>,
+    prev: Cell<Option<&'a Syntax<'a>>>,
+    parent: Cell<Option<&'a Syntax<'a>>>,
+    prev_is_contiguous: Cell<bool>,
+    change: Cell<Option<ChangeKind<'a>>>,
+    num_ancestors: Cell<u32>,
+    unique_id: Cell<u32>,
+    content_id: Cell<u32>,
 }
 
 impl<'a> SyntaxInfo<'a> {
-    pub fn new(pos_content_hash: u64) -> Self {
+    pub fn new() -> Self {
         Self {
-            pos_content_hash,
             next: Cell::new(None),
             prev: Cell::new(None),
+            parent: Cell::new(None),
             prev_is_contiguous: Cell::new(false),
             change: Cell::new(None),
             num_ancestors: Cell::new(0),
             unique_id: Cell::new(0),
+            content_id: Cell::new(0),
         }
     }
 }
@@ -74,7 +75,7 @@ pub enum Syntax<'a> {
         children: Vec<&'a Syntax<'a>>,
         close_position: Vec<SingleLineSpan>,
         close_content: String,
-        num_descendants: u64,
+        num_descendants: u32,
     },
     Atom {
         info: SyntaxInfo<'a>,
@@ -107,7 +108,11 @@ impl<'a> fmt::Debug for Syntax<'a> {
                 info,
                 ..
             } => {
-                let mut ds = f.debug_struct(&format!("List id:{}", self.id()));
+                let mut ds = f.debug_struct(&format!(
+                    "List id:{} content:{}",
+                    self.id(),
+                    self.content_id()
+                ));
 
                 ds.field("open_content", &open_content)
                     .field("open_position", &dbg_pos(open_position))
@@ -134,7 +139,11 @@ impl<'a> fmt::Debug for Syntax<'a> {
                 info,
                 ..
             } => {
-                let mut ds = f.debug_struct(&format!("Atom id:{}", self.id()));
+                let mut ds = f.debug_struct(&format!(
+                    "Atom id:{} content:{}",
+                    self.id(),
+                    self.content_id()
+                ));
                 ds.field("content", &content);
                 ds.field("position", &dbg_pos(position));
 
@@ -173,18 +182,8 @@ impl<'a> Syntax<'a> {
             };
         }
 
-        let mut hasher = DefaultHasher::new();
-
-        open_position.hash(&mut hasher);
-        open_content.hash(&mut hasher);
-        close_content.hash(&mut hasher);
-        close_position.hash(&mut hasher);
-        for child in &children {
-            child.hash(&mut hasher);
-        }
-
         arena.alloc(List {
-            info: SyntaxInfo::new(hasher.finish()),
+            info: SyntaxInfo::new(),
             open_position,
             open_content: open_content.into(),
             close_content: close_content.into(),
@@ -217,13 +216,8 @@ impl<'a> Syntax<'a> {
         content: &str,
         is_comment: bool,
     ) -> &'a mut Syntax<'a> {
-        let mut hasher = DefaultHasher::new();
-
-        position.hash(&mut hasher);
-        content.hash(&mut hasher);
-
         arena.alloc(Atom {
-            info: SyntaxInfo::new(hasher.finish()),
+            info: SyntaxInfo::new(),
             position,
             content: content.into(),
             is_comment,
@@ -244,14 +238,27 @@ impl<'a> Syntax<'a> {
         self.info().prev_is_contiguous.get()
     }
 
-    pub fn id(&self) -> u64 {
+    /// A unique ID of this syntax node. Every node is guaranteed to
+    /// have a different value.
+    pub fn id(&self) -> u32 {
         self.info().unique_id.get()
     }
 
-    pub fn num_ancestors(&self) -> u64 {
+    /// A content ID of this syntax node. Two nodes have the same
+    /// content ID if they have the same content, regardless of
+    /// position.
+    fn content_id(&self) -> u32 {
+        self.info().content_id.get()
+    }
+
+    pub fn num_ancestors(&self) -> u32 {
         self.info().num_ancestors.get()
     }
 
+    pub fn change(&'a self) -> Option<ChangeKind<'a>> {
+        self.info().change.get()
+    }
+
     pub fn first_line(&self) -> Option<LineNumber> {
         let position = match self {
             List { open_position, .. } => open_position,
@@ -293,147 +300,88 @@ impl<'a> Syntax<'a> {
             };
         }
     }
-
-    pub fn equal_content(&self, other: &Self) -> bool {
-        match (&self, other) {
-            (
-                Atom {
-                    content: lhs_content,
-                    is_comment: lhs_is_comment,
-                    ..
-                },
-                Atom {
-                    content: rhs_content,
-                    is_comment: rhs_is_comment,
-                    ..
-                },
-            ) => {
-                if lhs_is_comment != rhs_is_comment {
-                    return false;
-                }
-
-                if *lhs_is_comment {
-                    let is_multiline = lhs_content.lines().count() > 1;
-
-                    if is_multiline {
-                        let lhs_lines = lhs_content.lines().map(|l| l.trim_start()).collect_vec();
-                        let rhs_lines = rhs_content.lines().map(|l| l.trim_start()).collect_vec();
-                        return lhs_lines == rhs_lines;
-                    }
-                }
-                lhs_content == rhs_content
-            }
-            (
-                List {
-                    open_content: lhs_open_content,
-                    close_content: lhs_close_content,
-                    children: lhs_children,
-                    ..
-                },
-                List {
-                    open_content: rhs_open_content,
-                    close_content: rhs_close_content,
-                    children: rhs_children,
-                    ..
-                },
-            ) => {
-                if lhs_open_content != rhs_open_content || lhs_close_content != rhs_close_content {
-                    return false;
-                }
-                if lhs_children.len() != rhs_children.len() {
-                    return false;
-                }
-                for (lhs_child, rhs_child) in lhs_children.iter().zip(rhs_children.iter()) {
-                    if !lhs_child.equal_content(rhs_child) {
-                        return false;
-                    }
-                }
-
-                true
-            }
-            _ => false,
-        }
-    }
-
-    /// Does this `Node` have the same position in all its subnodes?
-    ///
-    /// Nodes with different numbers of children return false
-    /// regardless of top-level positions.
-    fn equal_pos(&self, other: &Self) -> bool {
-        match (&self, other) {
-            (
-                Atom {
-                    position: lhs_position,
-                    ..
-                },
-                Atom {
-                    position: rhs_position,
-                    ..
-                },
-            ) => lhs_position == rhs_position,
-            (
-                List {
-                    open_position: lhs_open_position,
-                    close_position: lhs_close_position,
-                    children: lhs_children,
-                    ..
-                },
-                List {
-                    open_position: rhs_open_position,
-                    close_position: rhs_close_position,
-                    children: rhs_children,
-                    ..
-                },
-            ) => {
-                if lhs_open_position != rhs_open_position
-                    || lhs_close_position != rhs_close_position
-                {
-                    return false;
-                }
-                if lhs_children.len() != rhs_children.len() {
-                    return false;
-                }
-
-                for (lhs_child, rhs_child) in lhs_children.iter().zip(rhs_children.iter()) {
-                    if !lhs_child.equal_pos(rhs_child) {
-                        return false;
-                    }
-                }
-
-                true
-            }
-            _ => false,
-        }
-    }
-}
-
-pub fn init_info<'a>(lhs_roots: &[&'a Syntax<'a>], rhs_roots: &[&'a Syntax<'a>]) {
-    let next_id = init_info_single(lhs_roots, 0);
-    init_info_single(rhs_roots, next_id);
 }
 
 /// Initialise all the fields in `SyntaxInfo`.
-///
-/// Return the next unique ID available, so we can ensure LHS and RHS
-/// have different IDs.
-pub fn init_info_single<'a>(roots: &[&'a Syntax<'a>], first_id: u64) -> u64 {
-    set_next(roots, None);
-    set_prev(roots, None);
-    set_num_ancestors(roots, 0);
-    set_prev_is_contiguous(roots);
-    set_unique_id(roots, first_id)
+pub fn init_info<'a>(lhs_roots: &[&'a Syntax<'a>], rhs_roots: &[&'a Syntax<'a>]) {
+    let mut id = 0;
+    init_info_single(lhs_roots, &mut id);
+    init_info_single(rhs_roots, &mut id);
+
+    let mut existing = HashMap::new();
+    set_content_id(lhs_roots, &mut existing);
+    set_content_id(rhs_roots, &mut existing);
 }
 
-fn set_unique_id<'a>(nodes: &[&'a Syntax<'a>], prev_id: u64) -> u64 {
-    let mut id = prev_id + 1;
+type ContentKey = (Option<String>, Option<String>, Vec<u32>, bool, bool);
+
+fn set_content_id<'a>(nodes: &[&'a Syntax<'a>], existing: &mut HashMap<ContentKey, u32>) {
     for node in nodes {
-        node.info().unique_id.set(id);
+        let key: ContentKey = match node {
+            List {
+                open_content,
+                close_content,
+                children,
+                ..
+            } => {
+                // Recurse first, so children all have their content_id set.
+                set_content_id(children, existing);
+
+                let children_content_ids: Vec<_> =
+                    children.iter().map(|c| c.info().content_id.get()).collect();
+
+                (
+                    Some(open_content.clone()),
+                    Some(close_content.clone()),
+                    children_content_ids,
+                    true,
+                    false,
+                )
+            }
+            Atom {
+                content,
+                is_comment,
+                ..
+            } => {
+                let clean_content = if *is_comment && content.lines().count() > 1 {
+                    content
+                        .lines()
+                        .map(|l| l.trim_start())
+                        .collect::<Vec<_>>()
+                        .join("\n")
+                        .to_string()
+                } else {
+                    content.clone()
+                };
+                (Some(clean_content), None, vec![], false, *is_comment)
+            }
+        };
+
+        // Ensure the ID is always greater than zero, so we can
+        // distinguish an uninitialised SyntaxInfo value.
+        let next_id = existing.len() as u32 + 1;
+        let content_id = existing.entry(key).or_insert(next_id);
+        node.info().content_id.set(*content_id);
+    }
+}
+
+fn init_info_single<'a>(roots: &[&'a Syntax<'a>], next_id: &mut u32) {
+    set_next(roots, None);
+    set_prev(roots, None);
+    set_parent(roots, None);
+    set_num_ancestors(roots, 0);
+    set_prev_is_contiguous(roots);
+    set_unique_id(roots, next_id)
+}
+
+fn set_unique_id<'a>(nodes: &[&'a Syntax<'a>], next_id: &mut u32) {
+    for node in nodes {
+        node.info().unique_id.set(*next_id);
+        *next_id += 1;
         if let List { children, .. } = node {
-            id = set_unique_id(children, id);
+            set_unique_id(children, next_id);
         }
-        id += 1;
     }
-    id
 }
 
 /// For every syntax node in the tree, mark the next node according to
@@ -465,7 +413,16 @@ fn set_prev<'a>(nodes: &[&'a Syntax<'a>], parent: Option<&'a Syntax<'a>>) {
     }
 }
 
-fn set_num_ancestors<'a>(nodes: &[&Syntax<'a>], num_ancestors: u64) {
+fn set_parent<'a>(nodes: &[&'a Syntax<'a>], parent: Option<&'a Syntax<'a>>) {
+    for node in nodes {
+        node.info().parent.set(parent);
+        if let List { children, .. } = node {
+            set_parent(children, Some(node));
+        }
+    }
+}
+
+fn set_num_ancestors<'a>(nodes: &[&Syntax<'a>], num_ancestors: u32) {
     for node in nodes {
         node.info().num_ancestors.set(num_ancestors);
 
@@ -506,14 +463,16 @@ fn set_prev_is_contiguous<'a>(roots: &[&Syntax<'a>]) {
 
 impl<'a> PartialEq for Syntax<'a> {
     fn eq(&self, other: &Self) -> bool {
-        self.equal_pos(other) && self.equal_content(other)
+        debug_assert!(self.content_id() > 0);
+        debug_assert!(other.content_id() > 0);
+        self.content_id() == other.content_id()
     }
 }
 impl<'a> Eq for Syntax<'a> {}
 
 impl<'a> Hash for Syntax<'a> {
     fn hash<H: Hasher>(&self, state: &mut H) {
-        self.info().pos_content_hash.hash(state);
+        self.content_id().hash(state);
     }
 }
 
@@ -1121,6 +1080,7 @@ mod tests {
 
         let comment = Syntax::new_comment(&arena, pos.clone(), "foo");
         let atom = Syntax::new_atom(&arena, pos, "foo");
+        init_info(&[comment], &[atom]);
 
         assert_ne!(comment, atom);
     }
@@ -1137,41 +1097,43 @@ mod tests {
 
         let x = Syntax::new_comment(&arena, pos.clone(), "foo\nbar");
         let y = Syntax::new_comment(&arena, pos, "foo\n    bar");
+        init_info(&[x], &[y]);
 
         assert_eq!(x, y);
     }
 
     #[test]
     fn test_atom_equality_ignores_change() {
-        assert_eq!(
-            Atom {
-                info: SyntaxInfo {
-                    change: Cell::new(Some(Novel)),
-                    ..SyntaxInfo::new(1)
-                },
-
-                position: vec![SingleLineSpan {
-                    line: 1.into(),
-                    start_col: 2,
-                    end_col: 3
-                }],
-                content: "foo".into(),
-                is_comment: false,
+        let lhs = Atom {
+            info: SyntaxInfo {
+                change: Cell::new(Some(Novel)),
+                ..SyntaxInfo::new()
             },
-            Atom {
-                info: SyntaxInfo {
-                    change: Cell::new(None),
-                    ..SyntaxInfo::new(1)
-                },
-                position: vec![SingleLineSpan {
-                    line: 1.into(),
-                    start_col: 2,
-                    end_col: 3
-                }],
-                content: "foo".into(),
-                is_comment: false,
-            }
-        );
+
+            position: vec![SingleLineSpan {
+                line: 1.into(),
+                start_col: 2,
+                end_col: 3,
+            }],
+            content: "foo".into(),
+            is_comment: false,
+        };
+        let rhs = Atom {
+            info: SyntaxInfo {
+                change: Cell::new(None),
+                ..SyntaxInfo::new()
+            },
+            position: vec![SingleLineSpan {
+                line: 1.into(),
+                start_col: 2,
+                end_col: 3,
+            }],
+            content: "foo".into(),
+            is_comment: false,
+        };
+        init_info(&[&lhs], &[&rhs]);
+
+        assert_eq!(lhs, rhs);
     }
 
     #[test]
